/* titel: FILE.h, Implementation von getc(), putc(), fopen() nach
   K&R 2, p. 175ff. */

#include <stdlib.h>     /* Fuer malloc() */

#define EOF	(-1)
#define BUFSIZE	1024 	/* Zum Testen der Puffer-Funktionen nimmt man hier
			   besser einen kleinen Wert, etwa 32. */
#define OPEN_MAX 20     /* Maximalzahl gleichzeitig offener Dateien */

typedef struct _iobuf {
    int		cnt;	/* noch zu verarbeitende Zeichenpositionen */
    char*	ptr;	/* naechste freie Position */
    char*	base;	/* Pufferanfang */
    int		flag;	/* Zugriffsmodus */
    int		fd;	/* Filedeskriptor */
} FILE;

extern FILE _iob_[OPEN_MAX];

#define stdin	(&_iob_[0])
#define stdout	(&_iob_[1])
#define stderr	(&_iob_[2])

enum _flags {  /* Hier werden Oktalzahlen definiert: "Fuehrende 0" */
    _READ	= 01,	/* Leseberechtigung,       oktal-binaer: 000 001 */
    _WRITE	= 02,	/* Schreibberechtigung,    oktal-binaer: 000 010 */
    _UNBUF	= 04,	/* Ungepuffertes File,     oktal-binaer: 000 100 */
    _EOF	= 010,	/* EOF ist aufgetreten,    oktal-binaer: 001 000 */
    _ERR	= 020,  /* Fehler ist aufgetreten, oktal-binaer: 010 000 */
};

int _fillbuf(FILE*);
int _flushbuf(unsigned char , FILE*);

#define feof(p)		(((p)->flag & _EOF) != 0)
#define ferror(p)	(((p)->flag & _ERR) != 0)
#define fileno(p)	((p)->fd)

#define getc(p)		(--(p)->cnt >= 0 \
			 ? (unsigned char) *(p)->ptr++ : _fillbuf(p))
#define putc(x,p)	(--(p)->cnt >= 0 \
			 ? *(p)->ptr++ = (x) : _flushbuf((x),p))

#define getchar()	getc(stdin)
#define putchar(x)	putc((x),stdout)

#include <fcntl.h>	/* fuer O_READONLY usw. */
#define PERMS 	0666	/* oktal-binaer: 000 110 110 110 */

/* titel fopen: oeffnet File und gibt FILE* zurueck */

FILE *fopen(char* name, char* mode) {
    int fd;
    FILE* fp;
    
    if (*mode != 'r' && *mode != 'w' && *mode != 'a')
	return NULL;		/* Nur Lese-, Schreib- und Anhaeng-Modus */

    for (fp = _iob_; fp < _iob_ + OPEN_MAX; fp++)
	if ((fp->flag & (_READ | _WRITE)) == 0) /* oktal-binaer: 000 011 */
	    break;		/* Freies File gefunden */
    if (fp >= _iob_ + OPEN_MAX)	/* Kein freies File vorhanden */
	return NULL;

    if (*mode == 'w')
	fd = creat(name, PERMS);
    else if (*mode == 'a') {
	if ((fd = open(name, O_WRONLY, 0)) == -1)
	    fd = creat(name, PERMS);
	lseek(fd, 0L, 2);	/* lseek zum Aufsuchen der Schreibposition */
    } else
	fd = open(name, O_RDONLY, 0);
    if (fd == -1)
	return NULL;
    fp->fd = fd;
    fp->cnt = 0;
    fp->base = NULL;
    fp->flag = (*mode == 'r') ? _READ : _WRITE;
    return fp;
}

/* titel: _fillbuf: Allokiert und fuellt Eingabepuffer, gibt naechstes
   anliegendes  Zeichen zurueck, bzw. NULL im Fehler- oder EOF-Fall. */

int _fillbuf(FILE* fp) {
    int bufsize;
    
    if ((fp->flag & (_READ | _EOF | _ERR)) != _READ)	
	return EOF;			

    bufsize = (fp->flag  & _UNBUF) ? 1 : BUFSIZE;

    if (fp->base == NULL)
	if ((fp->base = (char*) malloc(bufsize)) == NULL)
	    return EOF;

    fp->ptr = fp->base;
    fp->cnt = read(fp->fd, fp->ptr, bufsize);
    if (--fp->cnt < 0) {
	if (fp->cnt == -1)
	    fp->flag |= _EOF;
	else
	    fp->flag |= _ERR;
	fp->cnt = 0;
	return EOF;
    }
    return (unsigned char) *fp->ptr++;
}

int _fflush_(FILE* fp) {	/* Schreibt alle Zeichen aus Puffer heraus */
    if ((fp->base == NULL)||((fp->flag & (_WRITE | _EOF | _ERR)) != _WRITE))
	return EOF;
    write(fp->fd, fp->base, fp->ptr - fp->base);
    fp->ptr = fp->base;
    fp->cnt = (fp->flag & _UNBUF) ? 1 : BUFSIZE;
    return 0;
}

/* Allokiert und leert Puffer, initialisiert neu und traegt x ein,
   x wird auch zurueckgegeben */
int _flushbuf(unsigned char x, FILE* fp) {  
    int bufsize;
    bufsize = (fp->flag & _UNBUF) ? 1 : BUFSIZE;
    if (fp->base == NULL) {
	if ((fp->base = (char*) malloc(bufsize)) == NULL)
	    return EOF;
	fp->ptr = fp->base;
	fp->cnt = bufsize; 
    }
    if ( _fflush_(fp) != 0 )
	return EOF;
    fp->cnt--;
    return (unsigned char) *fp->ptr++ = x;
}    

/* Standard-IO-Kanaele	*/

FILE _iob_[OPEN_MAX] = {
    { 0, (char*) 0, (char*) 0, _READ,           0},
    { 0, (char*) 0, (char*) 0, _WRITE,          1},
    { 0, (char*) 0, (char*) 0, _WRITE | _UNBUF, 2},
};
